"""Mork celery configuration file."""

from celery import Celery

app = Celery(
    "mork", include=["mork.celery.tasks.deletion", "mork.celery.tasks.emailing"]
)

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("mork.conf:settings", namespace="CELERY")
